# coding: utf-8
from __future__ import print_function

import json
from . import utils


class Message(object):
    def to_json(self):
        return json.dumps(self)


class PlainMessage(Message):
    def __init__(self, text):
        self.text = text

    def to_json(self):
        return json.dumps({
            'text': self.text,
        })


class BinaryMessage(Message):
    def __init__(self, bytes):
        self.bytes = bytes


class ExceptionMessage(Message):
    def __init__(self, exception):
        self.exception = exception


class JsonEventMessage(Message):
    def __init__(self, request_id, event_name, data, json):
        self.request_id = request_id
        self.event_name = event_name
        self.data = data
        self.json = json

    def to_json(self):
        return json.dumps({
            'requestId': self.request_id,
            'eventName': self.event_name,
            'data': self.data,
        })


class JsonMessage(Message):
    def __init__(self, json):
        self.json = json

    def to_json(self):
        return json.dumps(self.json)


class RpcMessage(Message):
    RPC_EVENT_NAME = '__rpc__'
    RPC_RESPONSE_EVENT_NAME = '__rpc_response__'

    def __init__(self, request_id, event_name, service_name, method_name, method_args, data, json):
        self.request_id = request_id
        self.event_name = event_name
        self.service_name = service_name
        self.method_name = method_name
        self.method_args = method_args
        self.data = data
        self.json = json

    def to_json(self):
        return json.dumps({
            'requestId': self.request_id,
            'eventName': self.event_name,
            'serviceName': self.service_name,
            'methodName': self.method_name,
            'methodArgs': self.method_args,
            'data': self.data,
        })


class ConnectMessageEvent(object):
    def __init__(self, channel):
        self.channel = channel


class ConnectMessage(Message):
    def __init__(self, channel):
        self.channel = channel


class CloseMessageEvent(object):
    def __init__(self, channel):
        self.channel = channel


class CloseMessage(Message):
    def __init__(self, channel):
        self.channel = channel


class MessageBuilder(object):
    def create_messages(self, channel, data):
        return [(channel, msg) for msg in self.direct_create_messages(data)]

    def direct_create_messages(self, data):
        msgs = []
        if data is None:
            return msgs
        if isinstance(data, CloseMessageEvent):
            msgs.append(CloseMessage(data.channel))
            return msgs
        if isinstance(data, ConnectMessageEvent):
            msgs.append(ConnectMessage(data.channel))
            return msgs
        if isinstance(data, BaseException):
            msgs.append(ExceptionMessage(data))
            return msgs
        if isinstance(data, basestring):
            msgs.append(PlainMessage(data))
            json_data = utils.try_parse_json(data)
            if json_data is None:
                return msgs
            msgs.append(JsonMessage(json_data))
            if json_data.get('requestId') and json_data.get('eventName'):
                request_id = json_data.get('requestId')
                event_name = json_data.get('eventName')
                data = json_data.get('data')
                msgs.append(JsonEventMessage(request_id, event_name, data, json_data))
                if RpcMessage.RPC_EVENT_NAME == event_name or RpcMessage.RPC_RESPONSE_EVENT_NAME == event_name:
                    service_name = json_data.get('serviceName')
                    method_name = json_data.get('methodName')
                    method_args = json_data.get('methodArgs')
                    if method_args is None:
                        method_args = []
                    msgs.append(
                        RpcMessage(request_id, event_name, service_name, method_name, method_args, data, json_data))
        # TODO: create binary type messages
        return msgs




# coding: utf-8
from __future__ import print_function
import threading
import time
from Queue import Queue


class Executor(object):
    def start(self):
        pass


class TaskPool(object):
    """TODO: concurrent queue"""

    def __init__(self):
        self._queue = Queue()

    def push(self, task):
        self._queue.put(task)

    def poll(self):
        if self._queue.qsize() > 0:
            try:
                return self._queue.get_nowait()
            except Exception as e:
                return None
        else:
            return None

    def is_empty(self):
        return self._queue.qsize() < 1

    def available(self):
        return self._queue.qsize() > 0


class TaskThread(threading.Thread):
    def __init__(self, task_pool):
        threading.Thread.__init__(self)
        self._task_pool = task_pool

    def run(self):
        while True:
            task = self._task_pool.poll()
            if task is not None:
                task()
            time.sleep(0.05)


class FixedThreadsExecutor(Executor):
    """TODO: check thread status and create new thread if some died"""

    def __init__(self, thread_count):
        self._thread_count = thread_count if thread_count > 0 else 1
        self._threads = []
        self._task_pool = TaskPool()

    def _init_threads(self):
        for i in range(self._thread_count):
            t = TaskThread(self._task_pool)
            self._threads.append(t)
            t.start()

    def start(self):
        self._init_threads()

    def submit(self, task):
        self._task_pool.push(task)


# coding: utf-8
from __future__ import print_function
import threading
import logging
from . import messages


class MessageProcessor(threading.Thread):
    def __init__(self, receive_msg_pool, send_msg_pool):
        threading.Thread.__init__(self)
        self.receive_msg_pool = receive_msg_pool
        self.send_msg_pool = send_msg_pool
        self._handlers = []
        self._handlers_map = {}
        self._init_handlers()

    def run(self):
        while True:
            pair = self.receive_msg_pool.take_message()
            if pair is not None:
                try:
                    self._process_message(pair[0], pair[1])
                except Exception as e:
                    print(e)
                    logging.error(e)

    def _init_handlers(self):
        from . import listeners

        self._handlers.append(listeners.PlainMessageHandler(self.send_msg_pool).listen(self.receive_msg_pool))
        self._handlers.append(listeners.BinaryMessageHandler(self.send_msg_pool).listen(self.receive_msg_pool))
        self._handlers.append(listeners.CloseMessageHandler(self.send_msg_pool).listen(self.receive_msg_pool))
        self._handlers.append(listeners.ExceptionMessageHandler(self.send_msg_pool).listen(self.receive_msg_pool))
        self._handlers.append(listeners.ConnectMessageHandler(self.send_msg_pool).listen(self.receive_msg_pool))
        self._handlers.append(listeners.JsonMessageHandler(self.send_msg_pool).listen(self.receive_msg_pool))
        self._handlers.append(listeners.JsonEventMessageHandler(self.send_msg_pool).listen(self.receive_msg_pool))
        self._handlers.append(listeners.RpcMessageHandler(self.send_msg_pool).listen(self.receive_msg_pool))
        for handler in self._handlers:
            self._handlers_map[handler.interest()] = handler

    def add_rpc_service(self, name, rpc_service):
        handler = self._handlers_map.get(messages.RpcMessage)
        if handler is not None:
            handler.add_rpc_service(name, rpc_service)

    def add_rpc_services(self, pairs):
        handler = self._handlers_map.get(messages.RpcMessage)
        if handler is not None:
            handler.add_rpc_services(pairs)

    def _process_message(self, channel, msg):
        handler = self._handlers_map.get(type(msg))
        if handler is not None:
            handler.process(channel, msg)
        else:
            print("Can't find handler for msg type %s" % str(type(msg)))


# coding: utf-8
from __future__ import print_function

from pyjznet import Server, rpc_method
import unittest
import logging


class DemoService(object):
    # @rpc_method
    def sayHi(self, name, remote_service):
        try:
            age = remote_service.call_rpc('demo', 'getAge')
        except Exception as e:
            logging.error(e)
            return str(e)
        logging.info('got say hi request of %s who\'s age is %d' % (name, age))
        return "Hi, %s who's age is %d!" % (name, age)


class TestNetServer(unittest.TestCase):
    def setUp(self):
        pass

    def test_server(self):
        server = Server()
        server.add_rpc_service('demo', DemoService())
        server.start()


if __name__ == '__main__':
    unittest.main()


# coding: utf-8
from __future__ import print_function
import json
import time
from datetime import datetime, timedelta


def try_parse_json(text):
    if text is None:
        return None
    try:
        return json.loads(text)
    except Exception as e:
        return None


def safeunicode(obj, encoding='utf-8'):
    r"""
    Converts any given object to unicode string.

        >>> safeunicode('hello')
        u'hello'
        >>> safeunicode(2)
        u'2'
        >>> safeunicode('\xe1\x88\xb4')
        u'\u1234'
    """
    t = type(obj)
    if t is unicode:
        return obj
    elif t is str:
        return obj.decode(encoding, 'ignore')
    elif t in [int, float, bool]:
        return unicode(obj)
    elif hasattr(obj, '__unicode__') or isinstance(obj, unicode):
        try:
            return unicode(obj)
        except Exception as e:
            return u""
    else:
        return str(obj).decode(encoding, 'ignore')


class TimeoutException(Exception):
    pass


class ValueFuture(object):
    def __init__(self):
        self.fulfill_callbacks = []
        self.reject_callbacks = []
        self.value = None
        self._is_done = False

    def done(self):
        self._is_done = True

    def is_done(self):
        return self._is_done

    def set(self, value):
        self.value = value

    def set_result(self, value):
        if self.is_done():
            return
        self.set(value)
        self.done()
        self.call_fulfills(value)

    def call_fulfills(self, value):
        for fulfill in self.fulfill_callbacks:
            if fulfill is not None:
                try:
                    fulfill(value)
                except Exception as e:
                    pass

    def call_rejects(self, exception):
        for reject in self.reject_callbacks:
            if reject is not None:
                try:
                    reject(exception)
                except Exception as e:
                    pass

    def cancel(self, may_interrupt_if_running=False):
        self._is_done = True
        self.call_rejects(Exception('future task canceled'))

    def get(self, timeout=5000):
        if self.is_done():
            return self.value
        start_time = datetime.now()

        def timedout():
            return start_time + timedelta(milliseconds=timeout) < datetime.now()

        while not self.is_done() and not timedout():
            time.sleep(0.05)
        if not self.is_done():
            raise TimeoutException()
        return self.value

    def fulfill(self, callback):
        if callback is not None:
            self.fulfill_callbacks.append(callback)
            if self.is_done():
                try:
                    callback(self.get())
                except Exception as e:
                    pass

    def reject(self, callback):
        if callback is not None:
            self.reject_callbacks.append(callback)


# coding: utf-8
from __future__ import print_function
import logging

from . import messages
from . import services


class PoolListener(object):
    def watch_pool(self, pool, key_type):
        pool.add_listener(key_type, self)

    def on_message(self, msg):
        pass

    def create_process_message_task(self, msg):
        return lambda: self.on_message(msg)


class MessageHandler(PoolListener):
    def __init__(self):
        self._send_msg_pool = None

    @property
    def send_msg_pool(self):
        return self._send_msg_pool

    def interest(self):
        pass

    def listen(self, pool):
        self.watch_pool(pool, self.interest())
        return self

    def process(self, channel, msg):
        pass

    def on_message(self, msg):
        self.process(msg[0], msg[1])


class SimpleMessageHandler(MessageHandler):
    def __init__(self, pool):
        MessageHandler.__init__(self)
        self._send_msg_pool = pool


class PlainMessageHandler(SimpleMessageHandler):
    def interest(self):
        return messages.PlainMessage

    def process(self, channel, msg):
        logging.info("got plain message %s" % str(msg.to_json()))


class BinaryMessageHandler(SimpleMessageHandler):
    def interest(self):
        return messages.BinaryMessage

    def process(self, channel, msg):
        print("got binary message", msg)


class CloseMessageHandler(SimpleMessageHandler):
    def interest(self):
        return messages.CloseMessage

    def process(self, channel, msg):
        print("got close message")


class ExceptionMessageHandler(SimpleMessageHandler):
    def interest(self):
        return messages.ExceptionMessage

    def process(self, channel, msg):
        print('got exception message', msg)


class ConnectMessageHandler(SimpleMessageHandler):
    def interest(self):
        return messages.ConnectMessage

    def process(self, channel, msg):
        print('got connect msg', msg)


class JsonMessageHandler(SimpleMessageHandler):
    def interest(self):
        return messages.JsonMessage

    def process(self, channel, msg):
        logging.info('got json message %s' % str(msg.to_json()))


class JsonEventMessageHandler(SimpleMessageHandler):
    def interest(self):
        return messages.JsonEventMessage

    def process(self, channel, msg):
        logging.info('got json event message %s' % str(msg.to_json()))


class SendMessageHandler(PoolListener):
    def __init__(self):
        PoolListener.__init__(self)

    def on_message(self, msg):
        print('need to send msg', msg)
        channel = msg[0]
        msg = msg[1]
        if isinstance(msg, messages.CloseMessage) or isinstance(msg, messages.ExceptionMessage):
            channel.ws.close()
        elif isinstance(msg, messages.PlainMessage) or isinstance(msg, messages.JsonMessage) \
                or isinstance(msg, messages.JsonEventMessage) or isinstance(msg, messages.RpcMessage):
            channel.ws.send(msg.to_json())
        else:
            pass  # TODO: binary message send


class RpcMessageHandler(SimpleMessageHandler):
    def __init__(self, pool):
        SimpleMessageHandler.__init__(self, pool)
        self._rpc_services = {}

    def add_rpc_service(self, name, rpc_service):
        self._rpc_services[name] = rpc_service

    def add_rpc_services(self, pairs):
        for pair in pairs:
            self.add_rpc_service(pair[0], pair[1])

    def interest(self):
        return messages.RpcMessage

    def process(self, channel, msg):
        logging.info('got rpc message %s' % str(msg.to_json()))
        try:
            event_name = msg.event_name
            if messages.RpcMessage.RPC_EVENT_NAME == event_name:
                self._process_rpc_request(channel, msg)
            elif messages.RpcMessage.RPC_RESPONSE_EVENT_NAME == event_name:
                self._process_rpc_response(channel, msg)
        except Exception as e:
            logging.error(e)

    def _find_method_of_rpc_service(self, rpc_service, method_name):
        methods = dir(rpc_service)
        for name in methods:
            method = getattr(rpc_service, name)
            if name == method_name:
                return method
                # attr_name = '__rpc_method_name_%s' % name
                # attr_val = getattr(rpc_service, attr_name)
                # if attr_val is not None:
                # if attr_val == method_name:
                # return method
        return None

    def _process_rpc_request(self, channel, msg):
        if channel is None or msg is None:
            return
        service_name = msg.service_name
        method_name = msg.method_name
        method_args = msg.method_args
        rpc_service = self._rpc_services.get(service_name)
        if rpc_service is None:
            logging.info("Can't find rpc service %s/%s" % (service_name, method_name))
            return
        method = self._find_method_of_rpc_service(rpc_service, method_name)
        if method is None:
            logging.info("Can't find rpc service %s/%s" % (service_name, method_name))
            return
        try:
            kwargs = {
                'remote_service': services.ServerRemoteService(channel, self.send_msg_pool)
            }
            res = method(*method_args, **kwargs)
        except TypeError as e:
            res = method(*method_args)
        msg = messages.RpcMessage(msg.request_id, messages.RpcMessage.RPC_RESPONSE_EVENT_NAME,
                                  service_name, method_name, None, res, None)
        self._send_msg_pool.offer_message((channel, msg))

    def _process_rpc_response(self, channel, msg):
        if channel is None or msg is None:
            return
        task_pair = services.RemoteService.rpc_task_pool.get(msg.request_id)
        if task_pair is None or task_pair[1] is None:
            logging.info("Can't find the request for this response: %s" % msg.request_id)
            return
        future = task_pair[1]
        future.set_result(msg.data)



# coding: utf-8
from __future__ import print_function

from . import pools
from . import messages
from . import id_helper
from . import utils


class RemoteService(object):
    rpc_task_pool = pools.RpcTaskPool(20000, lambda pair: pair[1].cancel(True) if pair and pair[1] else None)


class ServerRemoteService(RemoteService):
    def __init__(self, channel, send_msg_pool):
        self.channel = channel
        self.send_msg_pool = send_msg_pool

    def call_rpc_async(self, service_name, method_name, *method_args):
        msg = messages.RpcMessage(
            id_helper.generate_request_id(), messages.RpcMessage.RPC_EVENT_NAME,
            service_name, method_name, method_args, None, None)
        future = utils.ValueFuture()
        RemoteService.rpc_task_pool.put(msg.request_id, (self.channel, future))
        self.send_msg_pool.offer_message((self.channel, msg))
        return future

    def call_rpc(self, service_name, method_name, *method_args):
        try:
            return self.call_rpc_async(service_name, method_name, *method_args).get()
        except Exception as e:
            return e


# coding: utf-8
from __future__ import print_function
from geventwebsocket import WebSocketServer, WebSocketApplication, Resource
import sys
import logging
from . import pools
from . import processors
from . import listeners
from . import messages

root = logging.getLogger()
root.setLevel(logging.DEBUG)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
root.addHandler(ch)


class Server(object):
    def __init__(self):
        self.rpc_services = []
        self._send_msg_pool = pools.SendMessagePool()
        self._receive_msg_pool = pools.ReceiveMessagePool()
        self.message_builder = messages.MessageBuilder()

    def add_rpc_service(self, name, rpc_service):
        self.rpc_services.append((name, rpc_service))

    def start(self, host='127.0.0.1', port=8000):
        send_msg_pool = self._send_msg_pool
        receive_msg_pool = self._receive_msg_pool
        message_builder = self.message_builder
        send_msg_pool.add_listener(object, listeners.SendMessageHandler())

        class WSApplication(WebSocketApplication):
            def on_open(self, ):
                print("Connection opened")

            def on_close(self, reason):
                print('Connection closed')

            def on_message(self, message):
                print('got message', message)
                receive_msg_pool.offer_messages(message_builder.create_messages(self, message))
                # self.ws.send(message)

        msg_processor = processors.MessageProcessor(receive_msg_pool, send_msg_pool)
        msg_processor.add_rpc_services(self.rpc_services)
        msg_processor.start()
        ws_server = WebSocketServer(
            (host, port),
            Resource({'/': WSApplication,
                      '/websocket': WSApplication})
        )
        print("start to listen at http://%s:%d" % (host, port))
        ws_server.serve_forever()


import os
from setuptools import setup, find_packages


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()


setup(
    name='jznet',
    version='0.0.1',
    keywords=('net', 'websocket', 'RPC',),
    description='a python RPC library based on websocket',
    long_description=read('README.md'),
    license='MIT License',
    install_requires=['gevent', 'gevent-websocket'],
    author='zoowii',
    author_email='1992zhouwei@gmail.com',
    packages=find_packages(exclude=['*test']),
    platforms='any',
)


# coding: utf-8
from __future__ import print_function


class CoreService(object):
    """TODO"""
    CORE_SERVICE_NAME = '__core__'

    def getTags(self, remote_service):
        pass

    def addTag(self, tag, remote_service):
        pass

    def removeTag(self, tag, remote_service):
        pass

    def clearTags(self, remote_service):
        pass

    def containsTags(self, tags, remote_service):
        pass

    def getChannelId(self, remote_service):
        pass


# coding: utf-8
from __future__ import print_function
import logging
from Queue import Queue
import threading
import time
import functools
from . import threadpool
from datetime import datetime, timedelta

thread_pool = threadpool.FixedThreadsExecutor(4)
thread_pool.start()


def print_result(req, result):
    print("the result of %s is %s" % (str(req), str(result)))


class AbstractPool(object):
    def __init__(self):
        self._messages = Queue()
        self._listeners = {}

    def add_listener(self, interest, listener):
        self._listeners[interest] = listener

    def offer_message(self, msg):
        if msg is None:
            return
        listener = self._listeners.get(self.message_key_type(msg))
        if listener is not None:
            thread_pool.submit(listener.create_process_message_task(msg))
        other_listener = self._listeners.get(object)
        if other_listener is not None:
            thread_pool.submit(other_listener.create_process_message_task(msg))

    def offer_messages(self, msgs):
        for msg in msgs:
            self.offer_message(msg)

    def take_message(self):
        if self.available():
            return self._messages.get_nowait()
        else:
            return None

    def is_empty(self):
        return self._messages.qsize() < 1

    def available(self):
        return not self.is_empty()

    def message_key_type(self, msg):
        pass


class ChannelMessagePool(AbstractPool):
    def message_key_type(self, msg):
        return type(msg[1])


class ReceiveMessagePool(ChannelMessagePool):
    pass


class SendMessagePool(ChannelMessagePool):
    pass


class AutoClosePool(object):
    """TODO: concurrent"""

    def __init__(self, timeout_milliseconds=5000, close_fn=None):
        from .threadpool import FixedThreadsExecutor

        self.timeout_milliseconds = timeout_milliseconds
        self.close_fn = close_fn
        self.stopped = False
        self.close_lock = threading.Lock()
        self.pool_objects = {}
        self.executor = FixedThreadsExecutor(4)

        def check_close_task(pool):
            while pool.is_running():
                try:
                    pool.check_closed()
                except Exception as e:
                    pass
                time.sleep(1)

        self.executor.submit(functools.partial(check_close_task, self))
        self.executor.start()

    def is_running(self):
        return not self.stopped

    def stop(self):
        self.stopped = True

    def put(self, key, value):
        if key is not None:
            self.pool_objects[key] = (value, datetime.now())

    def remove(self, key):
        if key is None:
            return None
        pair = self.pool_objects.get(key)
        if pair is None:
            return None
        del self.pool_objects[key]
        return pair[0]

    def _is_available_pair(self, pair):
        if pair is None:
            return False
        update_time = pair[1]
        if update_time is None:
            return False
        expired_time = update_time + timedelta(milliseconds=self.timeout_milliseconds)
        if expired_time > datetime.now():
            return pair[0] is not None
        else:
            return False

    def get(self, key):
        pair = self.pool_objects.get(key, None)
        if pair is None:
            return None
        return pair[0] if self._is_available_pair(pair) else None

    def check_closed(self):
        for (key, pair) in self.pool_objects:
            if key is None:
                continue
            if not self._is_available_pair(pair):
                self.close_lock.acquire()
                try:
                    val = pair[0]
                    if self.close_fn is not None:
                        try:
                            self.close_fn(val)
                        except Exception as e:
                            pass
                    del self.pool_objects[key]
                finally:
                    self.close_lock.release()


class RpcTaskPool(AutoClosePool):
    pass


# coding: utf-8
from __future__ import print_function

import functools


def rpc_method(func, name=None):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        attr_name = '__rpc_method_name_%s' % func.__name__
        attr_value = name
        if name is None:
            attr_value = func.__name__
        setattr(func.__objclass__, attr_name, attr_value)
        return func(*args, **kwargs)

    return wrapped


